"""Standard packed message format classes."""

import json
import logging
from typing import List, Sequence, Tuple, Union

from ..core.profile import ProfileSession
from ..messaging.base_message import DIDCommVersion
from ..messaging.util import time_now
from ..protocols.routing.v1_0.messages.forward import Forward
from ..wallet.base import BaseWallet
from ..wallet.error import WalletError
from ..wallet.util import b64_to_str
from .error import RecipientKeysError, WireFormatEncodeError, WireFormatParseError
from .inbound.receipt import MessageReceipt
from .wire_format import BaseWireFormat

try:
    from .v2_pack_format import V2PackWireFormat
except ImportError:
    V2PackWireFormat = None

LOGGER = logging.getLogger(__name__)
DIDCOMM_V1_TYP = "JWM/1.0"
DIDCOMM_V2_TYP = "application/didcomm"
DIDCOMM_V1_ID = "@id"
DIDCOMM_V2_ID = "id"


def get_version_for_packed_msg(packed_msg: Union[str, bytes]):
    """Get the version of the packed message."""

    # Raise differnt errors? Not ValueError?
    protected_b64 = json.loads(packed_msg).get("protected")
    if not protected_b64:
        raise ValueError("Invalid message format")

    protected = json.loads(b64_to_str(protected_b64))

    typ = protected.get("typ")
    if not typ:
        raise ValueError("Unexpected protected headers format")

    # CONSTify the strings
    if DIDCOMM_V2_TYP in typ:
        return DIDCommVersion.v2

    if DIDCOMM_V1_TYP in typ:
        return DIDCommVersion.v1

    raise ValueError("Could not determine DIDComm version of packed message")


def get_version_for_outbound_msg(outbound_msg: Union[str, bytes]):
    """Get the version of the packed message."""

    msg_json = json.loads(outbound_msg)

    if DIDCOMM_V2_ID in msg_json:
        return DIDCommVersion.v2

    if DIDCOMM_V1_ID in msg_json:
        return DIDCommVersion.v1

    # Raise differnt errors? Not ValueError?
    raise ValueError("Could not determine DIDComm version of packed message")


class PackWireFormat(BaseWireFormat):
    """Standard DIDComm message parser and serializer."""

    def __init__(self):
        """Initialize the pack wire format instance."""
        self.v1pack_format = V1PackWireFormat()
        self.v2pack_format = V2PackWireFormat() if V2PackWireFormat is not None else None

    def get_for_packed_msg(self, packed_msg: Union[str, bytes]) -> BaseWireFormat:
        """Retrieve appropriate DIDComm instance for a given packed message."""
        pack_format = {
            DIDCommVersion.v1: self.v1pack_format,
            DIDCommVersion.v2: self.v2pack_format,
        }[get_version_for_packed_msg(packed_msg)]
        assert format, (
            "self.v2_pack_format will be set when --experimental-didcomm-v2 is set"
        )
        return pack_format

    async def parse_message(
        self, session: ProfileSession, message_body: Union[str, bytes]
    ) -> Tuple[dict, MessageReceipt]:
        """Pass an incoming message to the appropriately versioned PackWireFormat."""

        if session.profile.settings.get("experiment.didcomm_v2"):
            try:
                pack_format = self.get_for_packed_msg(message_body)
            except ValueError as err:
                raise WireFormatParseError(
                    "Unable to determine appropriate WireFormat version"
                ) from err

        else:
            pack_format = self.v1pack_format

        return await pack_format.parse_message(session, message_body)

    def get_for_outbound_msg(self, outbound_msg: Union[str, bytes]) -> BaseWireFormat:
        """Retrieve appropriate DIDComm instance for a given packed message."""
        return {
            DIDCommVersion.v1: self.v1pack_format,
            DIDCommVersion.v2: self.v2pack_format,
        }[get_version_for_outbound_msg(outbound_msg)]

    async def encode_message(
        self,
        session: ProfileSession,
        message_json: Union[str, bytes],
        recipient_keys: Sequence[str],
        routing_keys: Sequence[str],
        sender_key: str,
    ) -> Union[str, bytes]:
        """Pass an incoming message to the appropriately versioned PackWireFormat."""

        if session.profile.settings.get("experiment.didcomm_v2"):
            try:
                pack_format = self.get_for_outbound_msg(message_json)
            except ValueError as err:
                raise WireFormatParseError(
                    "Unable to determine appropriate WireFormat version"
                ) from err
        else:
            pack_format = self.v1pack_format

        return await pack_format.encode_message(
            session,
            message_json,
            recipient_keys,
            routing_keys,
            sender_key,
        )

    def get_recipient_keys(self, message_body: Union[str, bytes]) -> List[str]:
        """Get all recipient keys from a wire message."""
        return self.v1pack_format.get_recipient_keys(message_body=message_body)

    async def unpack(
        self,
        session: ProfileSession,
        message_body: Union[str, bytes],
        receipt: MessageReceipt,
    ):
        """Look up the wallet instance and perform the message unpack."""

        return await self.v1pack_format.unpack(
            session=session, message_body=message_body, receipt=receipt
        )

    async def pack(
        self,
        session: ProfileSession,
        message_json: Union[str, bytes],
        recipient_keys: Sequence[str],
        routing_keys: Sequence[str],
        sender_key: str,
    ):
        """Look up the wallet instance and perform the message pack."""

        return await self.v1pack_format.pack(
            session=session,
            message_json=message_json,
            recipient_keys=recipient_keys,
            routing_keys=routing_keys,
            sender_key=sender_key,
        )


class V1PackWireFormat(BaseWireFormat):
    """DIDComm V1 message parser and serializer."""

    async def parse_message(
        self,
        session: ProfileSession,
        message_body: Union[str, bytes],
    ) -> Tuple[dict, MessageReceipt]:
        """Deserialize an incoming message and further populate the request context.

        Args:
            session: The profile session for providing wallet access
            message_body: The body of the message

        Returns:
            A tuple of the parsed message and a message receipt instance

        Raises:
            WireFormatParseError: If the JSON parsing failed
            WireFormatParseError: If a wallet is required but can't be located

        """

        receipt = MessageReceipt()
        receipt.in_time = time_now()
        receipt.raw_message = message_body
        receipt.didcomm_version = DIDCommVersion.v1

        message_dict = None
        message_json = message_body

        if not message_json:
            raise WireFormatParseError("Message body is empty")

        try:
            message_dict = json.loads(message_json)
        except ValueError:
            raise WireFormatParseError("Message JSON parsing failed")
        if not isinstance(message_dict, dict):
            raise WireFormatParseError("Message JSON result is not an object")

        # packed messages are detected by the absence of @type
        if "@type" not in message_dict:
            try:
                message_json = await self.unpack(session, message_body, receipt)
            except WireFormatParseError:
                LOGGER.debug("Message unpack failed, falling back to JSON")
            else:
                receipt.raw_message = message_json
                try:
                    message_dict = json.loads(message_json)
                except ValueError:
                    raise WireFormatParseError("Message JSON parsing failed")
                if not isinstance(message_dict, dict):
                    raise WireFormatParseError("Message JSON result is not an object")

        # parse thread ID
        thread_dec = message_dict.get("~thread")
        receipt.thread_id = (
            thread_dec and thread_dec.get("thid") or message_dict.get("@id")
        )
        receipt.parent_thread_id = thread_dec.get("pthid") if thread_dec else None

        # handle transport decorator
        transport_dec = message_dict.get("~transport")
        if transport_dec:
            receipt.direct_response_mode = transport_dec.get("return_route")

        LOGGER.debug(f"Expanded message: {message_dict}")

        return message_dict, receipt

    async def unpack(
        self,
        session: ProfileSession,
        message_body: Union[str, bytes],
        receipt: MessageReceipt,
    ):
        """Look up the wallet instance and perform the message unpack."""
        wallet = session.inject_or(BaseWallet)
        if not wallet:
            raise WireFormatParseError("Wallet not defined in profile session")

        try:
            unpacked = await wallet.unpack_message(message_body)
            (
                message_json,
                receipt.sender_verkey,
                receipt.recipient_verkey,
            ) = unpacked
            return message_json
        except WalletError as e:
            raise WireFormatParseError("Message unpack failed") from e

    async def encode_message(
        self,
        session: ProfileSession,
        message_json: Union[str, bytes],
        recipient_keys: Sequence[str],
        routing_keys: Sequence[str],
        sender_key: str,
    ) -> Union[str, bytes]:
        """Encode an outgoing message for transport.

        Args:
            session: The profile session for providing wallet access
            message_json: The message body to serialize
            recipient_keys: A sequence of recipient verkeys
            routing_keys: A sequence of routing verkeys
            sender_key: The verification key of the sending agent

        Returns:
            The encoded message

        Raises:
            MessageEncodeError: If the message could not be encoded

        """

        if sender_key and recipient_keys:
            message = await self.pack(
                session, message_json, recipient_keys, routing_keys, sender_key
            )
        else:
            message = message_json
        return message

    async def pack(
        self,
        session: ProfileSession,
        message_json: Union[str, bytes],
        recipient_keys: Sequence[str],
        routing_keys: Sequence[str],
        sender_key: str,
    ):
        """Look up the wallet instance and perform the message pack."""
        if not sender_key or not recipient_keys:
            raise WireFormatEncodeError("Cannot pack message without associated keys")

        wallet = session.inject_or(BaseWallet)
        if not wallet:
            raise WireFormatEncodeError("No wallet instance")

        try:
            message = await wallet.pack_message(message_json, recipient_keys, sender_key)
        except WalletError as e:
            raise WireFormatEncodeError("Message pack failed") from e

        if routing_keys:
            recip_keys = recipient_keys
            for router_key in routing_keys:
                message = json.loads(message.decode("utf-8"))
                fwd_msg = Forward(to=recip_keys[0], msg=message)
                # Forwards are anon packed
                recip_keys = [router_key]
                try:
                    message = await wallet.pack_message(fwd_msg.to_json(), recip_keys)
                except WalletError as e:
                    raise WireFormatEncodeError("Forward message pack failed") from e
        return message

    def get_recipient_keys(self, message_body: Union[str, bytes]) -> List[str]:
        """Get all recipient keys from a wire message.

        Args:
            message_body: The body of the message

        Returns:
            List of recipient keys from the message body

        Raises:
            RecipientKeysError: If the recipient keys could not be extracted

        """

        try:
            message_dict = json.loads(message_body)
            protected = json.loads(b64_to_str(message_dict["protected"], urlsafe=True))
            recipients = protected["recipients"]

            recipient_keys = [recipient["header"]["kid"] for recipient in recipients]
        except Exception as e:
            raise RecipientKeysError("Error trying to extract recipient keys from JWE", e)

        return recipient_keys



from unittest import TestCase

from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import ATTACHMENT_FORMAT, PRES_20_REQUEST
from ..pres_format import V20PresFormat
from ..pres_request import V20PresRequest

CD_ID = "NcYxiDXkpYi6ov5FcYDi1e:3:CL:12:tag1"
INDY_PROOF_REQ = [
    {
        "name": "proof-req",
        "version": "1.0",
        "nonce": "12345",
        "requested_attributes": {
            "0_player_uuid": {
                "name": "player",
                "restrictions": [
                    {
                        "cred_def_id": f"{CD_ID}",
                        "attr::player::value": "Richie Knucklez",
                    }
                ],
                "non_revoked": {
                    "from": 1234567890,
                    "to": 1234567890,
                },
            },
            "0_screencapture_uuid": {
                "name": "screenCapture",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
                "non_revoked": {
                    "from": 1234567890,
                    "to": 1234567890,
                },
            },
        },
        "requested_predicates": {
            "0_highscore_GE_uuid": {
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
                "non_revoked": {
                    "from": 1234567890,
                    "to": 1234567890,
                },
            }
        },
    },
    {
        "name": "proof-req",
        "version": "1.0",
        "nonce": "123456",
        "requested_attributes": {
            "0_player_uuid": {
                "name": "player",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            },
            "0_screencapture_uuid": {
                "name": "screenCapture",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            },
        },
        "requested_predicates": {
            "0_highscore_GE_uuid": {
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            }
        },
    },
    {
        "name": "proof-req",
        "version": "1.0",
        "nonce": "1234567",
        "requested_attributes": {},
        "requested_predicates": {
            "0_highscore_GE_uuid": {
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            }
        },
    },
]

PRES_REQ = [
    V20PresRequest(
        comment="Test",
        will_confirm=True,
        formats=[
            V20PresFormat(
                attach_id="indy",
                format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][V20PresFormat.Format.INDY.api],
            )
        ],
        request_presentations_attach=[
            AttachDecorator.data_base64(mapping=proof_req, ident="indy")
        ],
    )
    for proof_req in INDY_PROOF_REQ
]


class TestV20PresRequest(TestCase):
    """Presentation request tests."""

    def test_init_type(self):
        """Test initializer and type."""
        for i, pres_req in enumerate(PRES_REQ):
            assert pres_req.will_confirm
            assert len(pres_req.formats) == len(pres_req.request_presentations_attach)
            assert pres_req.request_presentations_attach[0].content == INDY_PROOF_REQ[i]
            assert pres_req.attachment(V20PresFormat.Format.INDY) == INDY_PROOF_REQ[i]
            assert pres_req._type == DIDCommPrefix.qualify_current(PRES_20_REQUEST)

    def test_attachment_no_target_format(self):
        """Test attachment behaviour for only unknown formats."""

        x_pres_req = V20PresRequest(
            comment="Test",
            formats=[V20PresFormat(attach_id="not_indy", format_="not_indy")],
            request_presentations_attach=[
                AttachDecorator.data_base64(
                    ident="not_indy", mapping=PRES_REQ[0].serialize()
                )
            ],
        )
        assert x_pres_req.attachment() is None

    def test_serde(self):
        """Test de/serialization."""
        for pres_req_msg in PRES_REQ:
            pres_req_dict = pres_req_msg.serialize()
            pres_req_obj = V20PresRequest.deserialize(pres_req_dict)
            assert type(pres_req_obj) is V20PresRequest

            pres_req_dict["request_presentations~attach"][0]["data"]["base64"] = (
                "eyJub3QiOiAiaW5keSJ9"
            )
            with self.assertRaises(BaseModelError):
                V20PresRequest.deserialize(pres_req_dict)

            pres_req_dict["request_presentations~attach"][0]["@id"] = "xxx"
            with self.assertRaises(BaseModelError):
                V20PresRequest.deserialize(pres_req_dict)

            pres_req_dict["request_presentations~attach"].append(
                {
                    "@id": "def",
                    "mime-type": "application/json",
                    "data": {"base64": "eyJub3QiOiAiaW5keSJ9"},
                }
            )  # more attachments than formats
            with self.assertRaises(BaseModelError):
                V20PresRequest.deserialize(pres_req_dict)

            pres_req_msg.formats.append(  # unknown format: no validation
                V20PresFormat(
                    attach_id="not_indy",
                    format_="not_indy",
                )
            )
            obj = pres_req_msg.serialize()
            obj["request_presentations~attach"].append(
                {
                    "@id": "not_indy",
                    "mime-type": "application/json",
                    "data": {"base64": "eyJub3QiOiAiaW5keSJ9"},
                }
            )
            V20PresRequest.deserialize(obj)



from unittest import IsolatedAsyncioTestCase

from marshmallow.exceptions import ValidationError

from ....admin.request_context import AdminRequestContext
from ....messaging.models.base import BaseModelError
from ....storage.error import StorageError, StorageNotFoundError
from ....tests import mock
from ....utils.testing import create_test_profile
from ....wallet.models.wallet_record import WalletRecord
from ...base import BaseMultitenantManager
from ...error import MultitenantManagerError
from .. import routes as test_module


class TestMultitenantRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={"wallet.type": "askar", "admin.admin_api_key": "secret-key"},
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }

        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_format_wallet_record_removes_wallet_key(self):
        wallet_record = WalletRecord(
            wallet_id="test",
            key_management_mode=WalletRecord.MODE_MANAGED,
            settings={
                "wallet.name": "wallet_name",
                "wallet.key": "wallet_key",
                "admin.admin_api_key": "secret-key",
            },
        )

        formatted = test_module.format_wallet_record(wallet_record)

        assert "wallet.key" not in formatted["settings"]

    async def test_wallets_list(self):
        with (
            mock.patch.object(
                test_module, "WalletRecord", autospec=True
            ) as mock_wallet_record,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            wallets = [
                mock.MagicMock(
                    serialize=mock.MagicMock(
                        return_value={
                            "wallet_id": "wallet_id",
                            "created_at": "1234567890",
                            "settings": {"wallet.name": "test"},
                        }
                    )
                ),
                mock.MagicMock(
                    serialize=mock.MagicMock(
                        return_value={
                            "wallet_id": "wallet_id",
                            "created_at": "1234567891",
                            "settings": {"wallet.name": "test"},
                        }
                    )
                ),
                mock.MagicMock(
                    serialize=mock.MagicMock(
                        return_value={
                            "wallet_id": "wallet_id",
                            "created_at": "1234567892",
                            "settings": {"wallet.name": "test"},
                        }
                    )
                ),
            ]
            mock_wallet_record.query = mock.CoroutineMock()
            mock_wallet_record.query.return_value = wallets

            await test_module.wallets_list(self.request)
            mock_response.assert_called_once_with(
                {"results": [test_module.format_wallet_record(w) for w in wallets]}
            )

    async def test_wallets_list_x(self):
        with mock.patch.object(
            test_module, "WalletRecord", autospec=True
        ) as mock_wallet_record:
            mock_wallet_record.query = mock.CoroutineMock()

            mock_wallet_record.query.side_effect = StorageError()
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.wallets_list(self.request)

            mock_wallet_record.query.side_effect = BaseModelError()
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.wallets_list(self.request)

    async def test_wallets_list_query(self):
        self.request.query = {"wallet_name": "test"}

        with (
            mock.patch.object(
                test_module, "WalletRecord", autospec=True
            ) as mock_wallet_record,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            wallets = [
                mock.MagicMock(
                    serialize=mock.MagicMock(
                        return_value={
                            "wallet_id": "wallet_id",
                            "created_at": "1234567890",
                            "settings": {"wallet.name": "test"},
                        }
                    )
                ),
            ]
            mock_wallet_record.query = mock.CoroutineMock()
            mock_wallet_record.query.return_value = wallets

            await test_module.wallets_list(self.request)
            mock_response.assert_called_once_with(
                {
                    "results": [
                        {
                            "wallet_id": "wallet_id",
                            "created_at": "1234567890",
                            "settings": {"wallet.name": "test"},
                        }
                    ]
                }
            )

    async def test_wallet_create_tenant_settings(self):
        body = {
            "wallet_name": "test",
            "default_label": "test_label",
            "wallet_type": "askar",
            "wallet_key": "test",
            "key_management_mode": "managed",
            "wallet_webhook_urls": [],
            "wallet_dispatch_type": "base",
            "extra_settings": {
                "ACAPY_LOG_LEVEL": "INFO",
                "ACAPY_INVITE_PUBLIC": True,
                "ACAPY_PUBLIC_INVITES": True,
            },
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        test_module.attempt_auto_author_with_endorser_setup = mock.CoroutineMock()

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            wallet_mock = mock.MagicMock(
                serialize=mock.MagicMock(
                    return_value={
                        "wallet_id": "test",
                        "settings": {},
                        "key_management_mode": body["key_management_mode"],
                    }
                )
            )  # wallet_record
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.create_auth_token = mock.CoroutineMock(
                return_value="test_token"
            )
            mock_multitenant_mgr.create_wallet = mock.CoroutineMock(
                return_value=wallet_mock
            )

            mock_multitenant_mgr.create_auth_token = mock.CoroutineMock(
                return_value="test_token"
            )
            mock_multitenant_mgr.get_wallet_profile = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_create(self.request)

            mock_multitenant_mgr.create_wallet.assert_called_once_with(
                {
                    "wallet.name": body["wallet_name"],
                    "wallet.type": body["wallet_type"],
                    "wallet.key": body["wallet_key"],
                    "wallet.webhook_urls": body["wallet_webhook_urls"],
                    "wallet.dispatch_type": body["wallet_dispatch_type"],
                    "log.level": "INFO",
                    "debug.invite_public": True,
                    "public_invites": True,
                },
                body["key_management_mode"],
            )
            mock_multitenant_mgr.create_auth_token.assert_called_once_with(
                wallet_mock, body["wallet_key"]
            )
            mock_response.assert_called_once_with(
                {**test_module.format_wallet_record(wallet_mock), "token": "test_token"}
            )
            assert mock_multitenant_mgr.get_wallet_profile.called
            assert test_module.attempt_auto_author_with_endorser_setup.called

    async def test_wallet_create(self):
        body = {
            "wallet_name": "test",
            "default_label": "test_label",
            "wallet_type": "askar",
            "wallet_key": "test",
            "key_management_mode": "managed",
            "wallet_webhook_urls": [],
            "wallet_dispatch_type": "base",
        }
        self.request.json = mock.CoroutineMock(return_value=body)
        test_module.attempt_auto_author_with_endorser_setup = mock.CoroutineMock()

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            wallet_mock = mock.MagicMock(
                serialize=mock.MagicMock(
                    return_value={
                        "wallet_id": "test",
                        "settings": {},
                        "key_management_mode": body["key_management_mode"],
                    }
                )
            )  # wallet_record
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.create_wallet = mock.CoroutineMock(
                return_value=wallet_mock
            )

            mock_multitenant_mgr.create_auth_token = mock.CoroutineMock(
                return_value="test_token"
            )
            mock_multitenant_mgr.get_wallet_profile = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_create(self.request)

            mock_multitenant_mgr.create_wallet.assert_called_once_with(
                {
                    "wallet.name": body["wallet_name"],
                    "wallet.type": body["wallet_type"],
                    "wallet.key": body["wallet_key"],
                    "wallet.webhook_urls": body["wallet_webhook_urls"],
                    "wallet.dispatch_type": body["wallet_dispatch_type"],
                },
                body["key_management_mode"],
            )
            mock_multitenant_mgr.create_auth_token.assert_called_once_with(
                wallet_mock, body["wallet_key"]
            )
            mock_response.assert_called_once_with(
                {
                    **test_module.format_wallet_record(wallet_mock),
                    "token": "test_token",
                }
            )
            assert mock_multitenant_mgr.get_wallet_profile.called
            assert test_module.attempt_auto_author_with_endorser_setup.called

    async def test_wallet_create_x(self):
        body = {}
        self.request.json = mock.CoroutineMock(return_value=body)
        mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
        mock_multitenant_mgr.create_wallet.side_effect = MultitenantManagerError()
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, mock_multitenant_mgr
        )

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_create(self.request)

    async def test_wallet_create_schema_validation_fails_indy_no_name_key(self):
        incorrect_body = {"wallet_type": "indy"}

        with self.assertRaises(ValidationError):
            schema = test_module.CreateWalletRequestSchema()
            schema.validate_fields(incorrect_body)

    async def test_wallet_create_optional_default_fields(self):
        body = {
            "wallet_name": "test",
            "wallet_key": "test",
            "wallet_key_derivation": "ARGON2I_MOD",
            "wallet_webhook_urls": [],
            "wallet_dispatch_type": "base",
            "label": "my_test_label",
            "image_url": "https://image.com",
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        with mock.patch.object(test_module.web, "json_response"):
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.create_wallet = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            mock_multitenant_mgr.create_auth_token = mock.CoroutineMock()
            mock_multitenant_mgr.get_wallet_profile = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )

            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_create(self.request)
            mock_multitenant_mgr.create_wallet.assert_called_once_with(
                {
                    "wallet.name": body["wallet_name"],
                    "wallet.type": "askar",
                    "wallet.key": body["wallet_key"],
                    "default_label": body["label"],
                    "image_url": body["image_url"],
                    "wallet.webhook_urls": body["wallet_webhook_urls"],
                    "wallet.dispatch_type": body["wallet_dispatch_type"],
                    "wallet.key_derivation_method": body["wallet_key_derivation"],
                },
                WalletRecord.MODE_MANAGED,
            )
            assert mock_multitenant_mgr.get_wallet_profile.called

    async def test_wallet_create_raw_key_derivation(self):
        body = {
            "wallet_name": "test",
            "wallet_key": "test",
            "wallet_key_derivation": "RAW",
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        with mock.patch.object(test_module.web, "json_response"):
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.create_wallet = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            mock_multitenant_mgr.create_auth_token = mock.CoroutineMock()
            mock_multitenant_mgr.get_wallet_profile = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )

            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_create(self.request)
            mock_multitenant_mgr.create_wallet.assert_called_once_with(
                {
                    "wallet.type": "askar",
                    "wallet.name": body["wallet_name"],
                    "wallet.key": body["wallet_key"],
                    "wallet.key_derivation_method": body["wallet_key_derivation"],
                    "wallet.webhook_urls": [],
                    "wallet.dispatch_type": "base",
                },
                WalletRecord.MODE_MANAGED,
            )
            assert mock_multitenant_mgr.get_wallet_profile.called

    async def test_wallet_update_tenant_settings(self):
        self.request.match_info = {"wallet_id": "test-wallet-id"}
        body = {
            "wallet_webhook_urls": ["test-webhook-url"],
            "wallet_dispatch_type": "default",
            "label": "test-label",
            "image_url": "test-image-url",
            "extra_settings": {
                "ACAPY_LOG_LEVEL": "INFO",
                "ACAPY_INVITE_PUBLIC": True,
                "ACAPY_PUBLIC_INVITES": True,
            },
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            settings = {
                "wallet.webhook_urls": body["wallet_webhook_urls"],
                "wallet.dispatch_type": body["wallet_dispatch_type"],
                "default_label": body["label"],
                "image_url": body["image_url"],
                "log.level": "INFO",
                "debug.invite_public": True,
                "public_invites": True,
            }
            wallet_mock = mock.MagicMock(
                serialize=mock.MagicMock(
                    return_value={
                        "wallet_id": "test-wallet-id",
                        "settings": settings,
                    }
                )
            )
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.update_wallet = mock.CoroutineMock(
                return_value=wallet_mock
            )

            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_update(self.request)

            mock_multitenant_mgr.update_wallet.assert_called_once_with(
                "test-wallet-id",
                settings,
            )
            mock_response.assert_called_once_with(
                {"wallet_id": "test-wallet-id", "settings": settings}
            )

    async def test_wallet_update(self):
        self.request.match_info = {"wallet_id": "test-wallet-id"}
        body = {
            "wallet_webhook_urls": ["test-webhook-url"],
            "wallet_dispatch_type": "default",
            "label": "test-label",
            "image_url": "test-image-url",
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            settings = {
                "wallet.webhook_urls": body["wallet_webhook_urls"],
                "wallet.dispatch_type": body["wallet_dispatch_type"],
                "default_label": body["label"],
                "image_url": body["image_url"],
            }
            wallet_mock = mock.MagicMock(
                serialize=mock.MagicMock(
                    return_value={
                        "wallet_id": "test-wallet-id",
                        "settings": settings,
                    }
                )
            )
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.update_wallet = mock.CoroutineMock(
                return_value=wallet_mock
            )

            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_update(self.request)

            mock_multitenant_mgr.update_wallet.assert_called_once_with(
                "test-wallet-id",
                settings,
            )
            mock_response.assert_called_once_with(
                {"wallet_id": "test-wallet-id", "settings": settings}
            )

    async def test_wallet_update_no_wallet_webhook_urls(self):
        self.request.match_info = {"wallet_id": "test-wallet-id"}
        body = {
            "label": "test-label",
            "image_url": "test-image-url",
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            settings = {
                "default_label": body["label"],
                "image_url": body["image_url"],
            }
            wallet_mock = mock.MagicMock(
                serialize=mock.MagicMock(
                    return_value={
                        "wallet_id": "test-wallet-id",
                        "settings": settings,
                    }
                )
            )
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.update_wallet = mock.CoroutineMock(
                return_value=wallet_mock
            )

            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_update(self.request)

            mock_multitenant_mgr.update_wallet.assert_called_once_with(
                "test-wallet-id",
                settings,
            )
            mock_response.assert_called_once_with(
                {"wallet_id": "test-wallet-id", "settings": settings}
            )

    async def test_wallet_update_empty_wallet_webhook_urls(self):
        self.request.match_info = {"wallet_id": "test-wallet-id"}
        body = {
            "wallet_webhook_urls": [],
            "label": "test-label",
            "image_url": "test-image-url",
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            settings = {
                "wallet.webhook_urls": [],
                "wallet.dispatch_type": "base",
                "default_label": body["label"],
                "image_url": body["image_url"],
            }
            wallet_mock = mock.MagicMock(
                serialize=mock.MagicMock(
                    return_value={
                        "wallet_id": "test-wallet-id",
                        "settings": settings,
                    }
                )
            )
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.update_wallet = mock.CoroutineMock(
                return_value=wallet_mock
            )

            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_update(self.request)

            mock_multitenant_mgr.update_wallet.assert_called_once_with(
                "test-wallet-id",
                settings,
            )
            mock_response.assert_called_once_with(
                {"wallet_id": "test-wallet-id", "settings": settings}
            )

    async def test_wallet_update_wallet_settings_x(self):
        self.request.match_info = {"wallet_id": "test-wallet-id"}
        body = {
            "wallet_webhook_urls": ["test-webhook-url"],
            "label": "test-label",
            "image_url": "test-image-url",
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        with mock.patch.object(test_module.web, "json_response"):
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.update_wallet = mock.CoroutineMock(
                side_effect=test_module.WalletSettingsError("bad settings")
            )

            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            with self.assertRaises(test_module.web.HTTPBadRequest) as context:
                await test_module.wallet_update(self.request)
            assert "bad settings" in str(context.exception)

    async def test_wallet_update_no_params(self):
        self.request.match_info = {"wallet_id": "test-wallet-id"}
        body = {}
        self.request.json = mock.CoroutineMock(return_value=body)

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_update(self.request)

    async def test_wallet_update_not_found(self):
        self.request.match_info = {"wallet_id": "test-wallet-id"}
        body = {"label": "test-label"}
        self.request.json = mock.CoroutineMock(return_value=body)
        mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
        mock_multitenant_mgr.update_wallet = mock.CoroutineMock(
            side_effect=StorageNotFoundError()
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, mock_multitenant_mgr
        )

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.wallet_update(self.request)

    async def test_wallet_get(self):
        self.request.match_info = {"wallet_id": "dummy"}
        mock_wallet_record = mock.MagicMock()
        mock_wallet_record.serialize = mock.MagicMock(
            return_value={"settings": {}, "wallet_id": "dummy"}
        )

        with (
            mock.patch.object(
                test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_wallet_record_retrieve_by_id,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_wallet_record_retrieve_by_id.return_value = mock_wallet_record

            await test_module.wallet_get(self.request)
            mock_response.assert_called_once_with({"settings": {}, "wallet_id": "dummy"})

    async def test_wallet_get_not_found(self):
        self.request.match_info = {"wallet_id": "dummy"}

        with mock.patch.object(
            test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_wallet_record_retrieve_by_id:
            mock_wallet_record_retrieve_by_id.side_effect = StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.wallet_get(self.request)

    async def test_wallet_get_x(self):
        self.request.match_info = {"wallet_id": "dummy"}
        mock_wallet_record = mock.MagicMock()
        mock_wallet_record.serialize = mock.MagicMock(
            side_effect=test_module.BaseModelError()
        )

        with mock.patch.object(
            test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_wallet_record_retrieve_by_id:
            mock_wallet_record_retrieve_by_id.return_value = mock_wallet_record

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.wallet_get(self.request)

    async def test_wallet_create_token_managed(self):
        self.request.has_body = False
        self.request.match_info = {"wallet_id": "dummy"}
        mock_wallet_record = mock.MagicMock()
        mock_wallet_record.serialize = mock.MagicMock(
            return_value={"settings": {}, "wallet_id": "dummy"}
        )

        with (
            mock.patch.object(
                test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_wallet_record_retrieve_by_id,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_wallet_record_retrieve_by_id.return_value = mock_wallet_record
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.create_auth_token = mock.CoroutineMock(
                return_value="test_token"
            )
            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_create_token(self.request)

            mock_multitenant_mgr.create_auth_token.assert_called_once_with(
                mock_wallet_record, None
            )
            mock_response.assert_called_once_with({"token": "test_token"})

    async def test_wallet_create_token_unmanaged(self):
        self.request.match_info = {"wallet_id": "dummy"}
        self.request.json = mock.CoroutineMock(return_value={"wallet_key": "dummy_key"})
        mock_wallet_record = mock.MagicMock()
        mock_wallet_record.serialize = mock.MagicMock(
            return_value={"settings": {}, "wallet_id": "dummy"}
        )

        with (
            mock.patch.object(
                test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_wallet_record_retrieve_by_id,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_wallet_record_retrieve_by_id.return_value = mock_wallet_record
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.create_auth_token = mock.CoroutineMock(
                return_value="test_token"
            )
            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )

            await test_module.wallet_create_token(self.request)

            mock_multitenant_mgr.create_auth_token.assert_called_once_with(
                mock_wallet_record, "dummy_key"
            )
            mock_response.assert_called_once_with({"token": "test_token"})

    async def test_wallet_create_token_managed_wallet_key_provided_throws(self):
        self.request.match_info = {"wallet_id": "dummy"}
        self.request.json = mock.CoroutineMock(return_value={"wallet_key": "dummy_key"})
        mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, mock_multitenant_mgr
        )
        mock_wallet_record = mock.MagicMock()
        mock_wallet_record.serialize = mock.MagicMock(
            return_value={"settings": {}, "wallet_id": "dummy"}
        )
        mock_wallet_record.requires_external_key = False

        with mock.patch.object(
            test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_wallet_record_retrieve_by_id:
            mock_wallet_record_retrieve_by_id.return_value = mock_wallet_record

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.wallet_create_token(self.request)

    async def test_wallet_create_token_x(self):
        self.request.has_body = False
        self.request.match_info = {"wallet_id": "dummy"}
        mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, mock_multitenant_mgr
        )

        with mock.patch.object(
            test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_wallet_record_retrieve_by_id:
            mock_wallet_record_retrieve_by_id.return_value = mock.MagicMock()

            with self.assertRaises(test_module.web.HTTPUnauthorized):
                mock_wallet_record_retrieve_by_id.side_effect = (
                    test_module.WalletKeyMissingError()
                )
                await test_module.wallet_create_token(self.request)

            with self.assertRaises(test_module.web.HTTPNotFound):
                mock_wallet_record_retrieve_by_id.side_effect = (
                    test_module.StorageNotFoundError()
                )
                await test_module.wallet_create_token(self.request)

    async def test_wallet_remove_managed(self):
        self.request.has_body = False
        self.request.match_info = {"wallet_id": "dummy"}
        mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
        mock_multitenant_mgr.remove_wallet = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, mock_multitenant_mgr
        )

        with (
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(
                test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
            ),
        ):
            result = await test_module.wallet_remove(self.request)

            mock_multitenant_mgr.remove_wallet.assert_called_once_with("dummy", None)
            mock_response.assert_called_once_with({})
            assert result == mock_response.return_value

    async def test_wallet_remove_unmanaged(self):
        self.request.match_info = {"wallet_id": "dummy"}
        self.request.json = mock.CoroutineMock(return_value={"wallet_key": "dummy_key"})
        mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
        mock_multitenant_mgr.remove_wallet = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, mock_multitenant_mgr
        )
        with (
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(
                test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
            ),
        ):
            result = await test_module.wallet_remove(self.request)

            mock_multitenant_mgr.remove_wallet.assert_called_once_with(
                "dummy", "dummy_key"
            )
            mock_response.assert_called_once_with({})
            assert result == mock_response.return_value

    async def test_wallet_remove_managed_wallet_key_provided_throws(self):
        self.request.match_info = {"wallet_id": "dummy"}
        self.request.json = mock.CoroutineMock(return_value={"wallet_key": "dummy_key"})

        mock_wallet_record = mock.MagicMock()
        mock_wallet_record.requires_external_key = False
        mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, mock_multitenant_mgr
        )

        with mock.patch.object(
            test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_wallet_record_retrieve_by_id:
            mock_wallet_record_retrieve_by_id.return_value = mock_wallet_record

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.wallet_remove(self.request)

    async def test_wallet_remove_x(self):
        self.request.has_body = False
        self.request.match_info = {"wallet_id": "dummy"}

        with mock.patch.object(
            test_module.WalletRecord, "retrieve_by_id", mock.CoroutineMock()
        ):
            mock_multitenant_mgr = mock.AsyncMock(BaseMultitenantManager, autospec=True)
            mock_multitenant_mgr.remove_wallet = mock.CoroutineMock()
            self.profile.context.injector.bind_instance(
                BaseMultitenantManager, mock_multitenant_mgr
            )
            with self.assertRaises(test_module.web.HTTPUnauthorized):
                mock_multitenant_mgr.remove_wallet.side_effect = (
                    test_module.WalletKeyMissingError()
                )
                await test_module.wallet_remove(self.request)

            with self.assertRaises(test_module.web.HTTPNotFound):
                mock_multitenant_mgr.remove_wallet.side_effect = (
                    test_module.StorageNotFoundError()
                )
                await test_module.wallet_remove(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



from unittest import IsolatedAsyncioTestCase

import pytest

from ...tests import mock
from ...utils.testing import create_test_profile
from .. import revocation_setup as test_module
from ..events import (
    CredDefFinishedEvent,
    CredDefFinishedPayload,
    RevRegDefFinishedEvent,
    RevRegDefFinishedPayload,
)
from ..models.revocation import RevRegDef, RevRegDefValue
from ..revocation import AnonCredsRevocation


@pytest.mark.anoncreds
class TestAnonCredsRevocationSetup(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile(
            settings={
                "wallet-type": "askar-anoncreds",
                "tails_server_base_url": "http://tails-server.com",
            }
        )
        self.revocation_setup = test_module.DefaultRevocationSetup()

    @mock.patch.object(
        AnonCredsRevocation,
        "create_and_register_revocation_registry_definition",
        return_value=None,
    )
    async def test_on_cred_def_support_revocation_registers_revocation_def(
        self, mock_register_revocation_registry_definition
    ):
        event = CredDefFinishedEvent(
            CredDefFinishedPayload(
                schema_id="schema_id",
                cred_def_id="cred_def_id",
                issuer_id="issuer_id",
                support_revocation=True,
                max_cred_num=100,
                options={},
            )
        )
        await self.revocation_setup.on_cred_def(self.profile, event)

        assert mock_register_revocation_registry_definition.called

    @mock.patch.object(
        AnonCredsRevocation,
        "create_and_register_revocation_registry_definition",
        return_value=None,
    )
    async def test_on_cred_def_not_support_rev_option(
        self, mock_register_revocation_registry_definition
    ):
        event = CredDefFinishedEvent(
            CredDefFinishedPayload(
                schema_id="schema_id",
                cred_def_id="cred_def_id",
                issuer_id="issuer_id",
                support_revocation=False,
                max_cred_num=100,
                options={},
            )
        )
        await self.revocation_setup.on_cred_def(self.profile, event)

        assert not mock_register_revocation_registry_definition.called

    @mock.patch.object(
        AnonCredsRevocation,
        "upload_tails_file",
        return_value=None,
    )
    @mock.patch.object(
        AnonCredsRevocation,
        "create_and_register_revocation_list",
        return_value=None,
    )
    @mock.patch.object(
        AnonCredsRevocation,
        "set_active_registry",
        return_value=None,
    )
    async def test_on_rev_reg_def_with_support_revoc_option_registers_list(
        self, mock_set_active_reg, mock_register, mock_upload
    ):
        event = RevRegDefFinishedEvent(
            RevRegDefFinishedPayload(
                rev_reg_def_id="rev_reg_def_id",
                rev_reg_def=RevRegDef(
                    tag="0",
                    cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                    value=RevRegDefValue(
                        max_cred_num=100,
                        public_keys={
                            "accum_key": {"z": "1 0BB...386"},
                        },
                        tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
                        tails_location="http://tails-server.com",
                    ),
                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                    type="CL_ACCUM",
                ),
                options={},
            )
        )

        await self.revocation_setup.on_rev_reg_def(self.profile, event)
        assert mock_upload.called
        assert mock_register.called
        assert mock_set_active_reg.called

    @mock.patch.object(
        AnonCredsRevocation,
        "upload_tails_file",
        return_value=None,
    )
    @mock.patch.object(
        AnonCredsRevocation,
        "create_and_register_revocation_list",
        return_value=None,
    )
    async def test_on_rev_reg_def_author_and_auto_create_rev_reg(
        self, mock_register, mock_upload
    ):
        self.profile.settings["endorser.author"] = True
        self.profile.settings["endorser.auto_create_rev_reg"] = True
        event = RevRegDefFinishedEvent(
            RevRegDefFinishedPayload(
                rev_reg_def_id="rev_reg_def_id",
                rev_reg_def=RevRegDef(
                    tag="tag",
                    cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                    value=RevRegDefValue(
                        max_cred_num=100,
                        public_keys={
                            "accum_key": {"z": "1 0BB...386"},
                        },
                        tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
                        tails_location="http://tails-server.com",
                    ),
                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                    type="CL_ACCUM",
                ),
                options={},
            )
        )

        await self.revocation_setup.on_rev_reg_def(self.profile, event)
        assert mock_upload.called
        assert mock_register.called

    @mock.patch.object(
        AnonCredsRevocation,
        "upload_tails_file",
        return_value=None,
    )
    @mock.patch.object(
        AnonCredsRevocation,
        "create_and_register_revocation_list",
        return_value=None,
    )
    async def test_on_rev_reg_def_author_and_do_not_auto_create_rev_reg(
        self, mock_register, mock_upload
    ):
        self.profile.settings["endorser.author"] = True
        self.profile.settings["endorser.auto_create_rev_reg"] = False
        event = RevRegDefFinishedEvent(
            RevRegDefFinishedPayload(
                rev_reg_def_id="rev_reg_def_id",
                rev_reg_def=RevRegDef(
                    tag="tag",
                    cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                    value=RevRegDefValue(
                        max_cred_num=100,
                        public_keys={
                            "accum_key": {"z": "1 0BB...386"},
                        },
                        tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
                        tails_location="http://tails-server.com",
                    ),
                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                    type="CL_ACCUM",
                ),
                options={},
            )
        )

        await self.revocation_setup.on_rev_reg_def(self.profile, event)
        assert not mock_upload.called
        assert not mock_register.called



"""Represents a DID exchange complete message under RFC 23."""

from marshmallow import EXCLUDE, ValidationError, pre_dump

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import DIDX_COMPLETE, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.complete_handler.DIDXCompleteHandler"


class DIDXComplete(AgentMessage):
    """Class representing a DID exchange completion."""

    class Meta:
        """Metadata for DID exchange completion."""

        handler_class = HANDLER_CLASS
        message_type = DIDX_COMPLETE
        schema_class = "DIDXCompleteSchema"

    def __init__(
        self,
        **kwargs,
    ):
        """Initialize DID exchange complete message object."""
        super().__init__(**kwargs)


class DIDXCompleteSchema(AgentMessageSchema):
    """DID exchange complete schema class."""

    class Meta:
        """DID exchange complete schema metadata."""

        model_class = DIDXComplete
        unknown = EXCLUDE

    @pre_dump
    def check_thread_deco(self, obj, **kwargs):
        """Thread decorator, and its thid and pthid, are mandatory."""
        if not obj._decorators.to_dict().get("~thread", {}).keys() >= {"thid", "pthid"}:
            raise ValidationError("Missing required field(s) in thread decorator")
        return obj



import json
from unittest import mock

import pytest
import pytest_asyncio
from aries_askar import AskarError, Key, KeyAlg, Session

from ....utils.jwe import JweEnvelope, JweRecipient, b64url
from ....utils.testing import create_test_profile
from .. import v2 as test_module

ALICE_KID = "did:example:alice#key-1"
BOB_KID = "did:example:bob#key-1"
CAROL_KID = "did:example:carol#key-2"
MESSAGE = b"Expecto patronum"


@pytest_asyncio.fixture
async def session():
    profile = await create_test_profile()
    async with profile.session() as session:
        yield session.handle
    del session
    await profile.close()


@pytest.mark.askar
class TestAskarDidCommV2:
    @pytest.mark.asyncio
    async def test_es_round_trip(self, session: Session):
        alg = KeyAlg.X25519
        bob_sk = Key.generate(alg)
        bob_pk = Key.from_jwk(bob_sk.get_jwk_public())
        carol_sk = Key.generate(KeyAlg.P256)  # testing mixed recipient key types
        carol_pk = Key.from_jwk(carol_sk.get_jwk_public())

        enc_message = test_module.ecdh_es_encrypt(
            {BOB_KID: bob_pk, CAROL_KID: carol_pk}, MESSAGE
        )

        # receiver must have the private keypair accessible
        await session.insert_key("my_sk", bob_sk, tags={"kid": BOB_KID})

        plaintext, recip_kid, sender_kid = await test_module.unpack_message(
            session, enc_message
        )
        assert recip_kid == BOB_KID
        assert sender_kid is None
        assert plaintext == MESSAGE

    @pytest.mark.asyncio
    async def test_es_encrypt_x(self):
        alg = KeyAlg.X25519
        bob_sk = Key.generate(alg)
        bob_pk = Key.from_jwk(bob_sk.get_jwk_public())

        with pytest.raises(
            test_module.DidcommEnvelopeError, match="No message recipients"
        ):
            _ = test_module.ecdh_es_encrypt({}, MESSAGE)

        with mock.patch(
            "aries_askar.Key.generate",
            mock.MagicMock(side_effect=AskarError(99, "")),
        ):
            with pytest.raises(
                test_module.DidcommEnvelopeError,
                match="Error creating content encryption key",
            ):
                _ = test_module.ecdh_es_encrypt({BOB_KID: bob_pk}, MESSAGE)

        with mock.patch(
            "aries_askar.Key.aead_encrypt",
            mock.MagicMock(side_effect=AskarError(99, "")),
        ):
            with pytest.raises(
                test_module.DidcommEnvelopeError,
                match="Error encrypting",
            ):
                _ = test_module.ecdh_es_encrypt({BOB_KID: bob_pk}, MESSAGE)

    @pytest.mark.asyncio
    async def test_es_decrypt_x(self):
        alg = KeyAlg.X25519
        bob_sk = Key.generate(alg)

        message_unknown_alg = JweEnvelope(
            protected={"alg": "NOT-SUPPORTED"},
        )
        message_unknown_alg.add_recipient(
            JweRecipient(encrypted_key=b"0000", header={"kid": BOB_KID})
        )
        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Unsupported ECDH-ES algorithm",
        ):
            _ = test_module.ecdh_es_decrypt(
                message_unknown_alg,
                BOB_KID,
                bob_sk,
            )

        message_unknown_enc = JweEnvelope(
            protected={"alg": "ECDH-ES+A128KW", "enc": "UNKNOWN"},
        )
        message_unknown_enc.add_recipient(
            JweRecipient(encrypted_key=b"0000", header={"kid": BOB_KID})
        )
        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Unsupported ECDH-ES content encryption",
        ):
            _ = test_module.ecdh_es_decrypt(
                message_unknown_enc,
                BOB_KID,
                bob_sk,
            )

        message_invalid_epk = JweEnvelope(
            protected={"alg": "ECDH-ES+A128KW", "enc": "A256GCM", "epk": {}},
        )
        message_invalid_epk.add_recipient(
            JweRecipient(encrypted_key=b"0000", header={"kid": BOB_KID})
        )
        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Error loading ephemeral key",
        ):
            _ = test_module.ecdh_es_decrypt(
                message_invalid_epk,
                BOB_KID,
                bob_sk,
            )

    @pytest.mark.asyncio
    async def test_1pu_round_trip(self, session: Session):
        alg = KeyAlg.X25519
        alice_sk = Key.generate(alg)
        alice_pk = Key.from_jwk(alice_sk.get_jwk_public())
        bob_sk = Key.generate(alg)
        bob_pk = Key.from_jwk(bob_sk.get_jwk_public())

        enc_message = test_module.ecdh_1pu_encrypt(
            {BOB_KID: bob_pk}, ALICE_KID, alice_sk, MESSAGE
        )

        # receiver must have the private keypair accessible
        await session.insert_key("my_sk", bob_sk, tags={"kid": BOB_KID})
        # for now at least, insert the sender public key so it can be resolved
        await session.insert_key("alice_pk", alice_pk, tags={"kid": ALICE_KID})

        plaintext, recip_kid, sender_kid = await test_module.unpack_message(
            session, enc_message
        )
        assert recip_kid == BOB_KID
        assert sender_kid == ALICE_KID
        assert plaintext == MESSAGE

    @pytest.mark.asyncio
    async def test_1pu_encrypt_x(self, session: Session):
        alg = KeyAlg.X25519
        alice_sk = Key.generate(alg)
        bob_sk = Key.generate(alg)
        bob_pk = Key.from_jwk(bob_sk.get_jwk_public())

        with pytest.raises(
            test_module.DidcommEnvelopeError, match="No message recipients"
        ):
            _ = test_module.ecdh_1pu_encrypt({}, ALICE_KID, alice_sk, MESSAGE)

        alt_sk = Key.generate(KeyAlg.P256)
        alt_pk = Key.from_jwk(alt_sk.get_jwk_public())
        with pytest.raises(
            test_module.DidcommEnvelopeError, match="key types must be consistent"
        ):
            _ = test_module.ecdh_1pu_encrypt(
                {BOB_KID: bob_pk, "alt": alt_pk}, ALICE_KID, alice_sk, MESSAGE
            )

        with mock.patch(
            "aries_askar.Key.generate",
            mock.MagicMock(side_effect=AskarError(99, "")),
        ):
            with pytest.raises(
                test_module.DidcommEnvelopeError,
                match="Error creating content encryption key",
            ):
                _ = test_module.ecdh_1pu_encrypt(
                    {BOB_KID: bob_pk}, ALICE_KID, alice_sk, MESSAGE
                )

        with mock.patch(
            "aries_askar.Key.aead_encrypt",
            mock.MagicMock(side_effect=AskarError(99, "")),
        ):
            with pytest.raises(
                test_module.DidcommEnvelopeError,
                match="Error encrypting",
            ):
                _ = test_module.ecdh_1pu_encrypt(
                    {BOB_KID: bob_pk}, ALICE_KID, alice_sk, MESSAGE
                )

    @pytest.mark.asyncio
    async def test_1pu_decrypt_x(self):
        alg = KeyAlg.X25519
        alice_sk = Key.generate(alg)
        alice_pk = Key.from_jwk(alice_sk.get_jwk_public())
        bob_sk = Key.generate(alg)

        message_unknown_alg = JweEnvelope(
            protected={"alg": "NOT-SUPPORTED"},
        )
        message_unknown_alg.add_recipient(
            JweRecipient(encrypted_key=b"0000", header={"kid": BOB_KID})
        )
        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Unsupported ECDH-1PU algorithm",
        ):
            _ = test_module.ecdh_1pu_decrypt(
                message_unknown_alg,
                BOB_KID,
                bob_sk,
                alice_pk,
            )

        message_unknown_enc = JweEnvelope(
            protected={"alg": "ECDH-1PU+A128KW", "enc": "UNKNOWN"},
        )
        message_unknown_enc.add_recipient(
            JweRecipient(encrypted_key=b"0000", header={"kid": BOB_KID})
        )
        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Unsupported ECDH-1PU content encryption",
        ):
            _ = test_module.ecdh_1pu_decrypt(
                message_unknown_enc, BOB_KID, bob_sk, alice_pk
            )

        message_invalid_epk = JweEnvelope(
            protected={"alg": "ECDH-1PU+A128KW", "enc": "A256CBC-HS512", "epk": {}},
        )
        message_invalid_epk.add_recipient(
            JweRecipient(encrypted_key=b"0000", header={"kid": BOB_KID})
        )
        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Error loading ephemeral key",
        ):
            _ = test_module.ecdh_1pu_decrypt(
                message_invalid_epk,
                BOB_KID,
                bob_sk,
                alice_pk,
            )

    @pytest.mark.asyncio
    async def test_unpack_message_any_x(self, session: Session):
        message_invalid = "{}"

        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Invalid",
        ):
            _ = await test_module.unpack_message(session, message_invalid)

        message_unknown_alg = json.dumps(
            {
                "protected": b64url(json.dumps({"alg": "NOT-SUPPORTED"})),
                "recipients": [{"header": {"kid": "bob"}, "encrypted_key": "MTIzNA"}],
                "iv": "MTIzNA",
                "ciphertext": "MTIzNA",
                "tag": "MTIzNA",
            }
        )

        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Unsupported DIDComm encryption",
        ):
            _ = await test_module.unpack_message(session, message_unknown_alg)

        message_unknown_recip = json.dumps(
            {
                "protected": b64url(json.dumps({"alg": "ECDH-ES+A128KW"})),
                "recipients": [{"header": {"kid": "bob"}, "encrypted_key": "MTIzNA"}],
                "iv": "MTIzNA",
                "ciphertext": "MTIzNA",
                "tag": "MTIzNA",
            }
        )

        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="No recognized recipient key",
        ):
            _ = await test_module.unpack_message(session, message_unknown_recip)

    @pytest.mark.asyncio
    async def test_unpack_message_1pu_x(self, session: Session):
        alg = KeyAlg.X25519
        alice_sk = Key.generate(alg)
        alice_pk = Key.from_jwk(alice_sk.get_jwk_public())
        bob_sk = Key.generate(alg)
        Key.from_jwk(bob_sk.get_jwk_public())

        # receiver must have the private keypair accessible
        await session.insert_key("my_sk", bob_sk, tags={"kid": BOB_KID})
        # for now at least, insert the sender public key so it can be resolved
        await session.insert_key("alice_pk", alice_pk, tags={"kid": ALICE_KID})

        message_1pu_no_skid = json.dumps(
            {
                "protected": b64url(json.dumps({"alg": "ECDH-1PU+A128KW"})),
                "recipients": [{"header": {"kid": BOB_KID}, "encrypted_key": "MTIzNA"}],
                "iv": "MTIzNA",
                "ciphertext": "MTIzNA",
                "tag": "MTIzNA",
            }
        )

        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Sender key ID not provided",
        ):
            _ = await test_module.unpack_message(session, message_1pu_no_skid)

        message_1pu_unknown_skid = json.dumps(
            {
                "protected": b64url(
                    json.dumps({"alg": "ECDH-1PU+A128KW", "skid": "UNKNOWN"})
                ),
                "recipients": [{"header": {"kid": BOB_KID}, "encrypted_key": "MTIzNA"}],
                "iv": "MTIzNA",
                "ciphertext": "MTIzNA",
                "tag": "MTIzNA",
            }
        )

        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Sender public key not found",
        ):
            _ = await test_module.unpack_message(session, message_1pu_unknown_skid)

        message_1pu_apu_invalid = json.dumps(
            {
                "protected": b64url(
                    json.dumps({"alg": "ECDH-1PU+A128KW", "skid": "A", "apu": "A"})
                ),
                "recipients": [{"header": {"kid": BOB_KID}, "encrypted_key": "MTIzNA"}],
                "iv": "MTIzNA",
                "ciphertext": "MTIzNA",
                "tag": "MTIzNA",
            }
        )

        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Invalid apu value",
        ):
            _ = await test_module.unpack_message(session, message_1pu_apu_invalid)

        message_1pu_apu_mismatch = json.dumps(
            {
                "protected": b64url(
                    json.dumps(
                        {
                            "alg": "ECDH-1PU+A128KW",
                            "skid": ALICE_KID,
                            "apu": b64url("UNKNOWN"),
                        }
                    )
                ),
                "recipients": [{"header": {"kid": BOB_KID}, "encrypted_key": "MTIzNA"}],
                "iv": "MTIzNA",
                "ciphertext": "MTIzNA",
                "tag": "MTIzNA",
            }
        )

        with pytest.raises(
            test_module.DidcommEnvelopeError,
            match="Mismatch between skid and apu",
        ):
            _ = await test_module.unpack_message(session, message_1pu_apu_mismatch)



from unittest import IsolatedAsyncioTestCase, mock

from .. import help as command


class TestHelp(IsolatedAsyncioTestCase):
    def test_exec_help(self):
        with (
            mock.patch.object(command.ArgumentParser, "print_help") as mock_print_help,
            mock.patch("builtins.print", mock.MagicMock()) as mock_print,
        ):
            command.execute([])
            mock_print_help.assert_called_once()

            command.execute(["-v"])
            mock_print.assert_called_once_with(command.__version__)

    def test_main(self):
        with (
            mock.patch.object(command, "__name__", "__main__") as mock_name,
            mock.patch.object(command, "execute", mock.MagicMock()) as mock_execute,
        ):
            command.main()
            mock_execute.assert_called_once



"""Credential offer message handler."""

from .....core.oob_processor import OobMessageProcessor
from .....indy.holder import IndyHolderError
from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError
from .....utils.tracing import get_timer, trace_event
from .....wallet.util import default_did_from_verkey
from .. import problem_report_for_record
from ..manager import CredentialManager, CredentialManagerError
from ..messages.credential_offer import CredentialOffer
from ..messages.credential_problem_report import ProblemReportReason


class CredentialOfferHandler(BaseHandler):
    """Message handler class for credential offers."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for credential offers.

        Args:
            context: request context
            responder: responder callback

        """
        r_time = get_timer()
        profile = context.profile
        self._logger.debug("CredentialOfferHandler called with context %s", context)
        assert isinstance(context.message, CredentialOffer)
        self._logger.debug(
            "Received credential offer message: %s",
            context.message.serialize(as_string=True),
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for credential offer not ready")

        # Find associated oob record
        oob_processor = context.inject(OobMessageProcessor)
        oob_record = await oob_processor.find_oob_record_for_inbound_message(context)

        # Either connection or oob context must be present
        if not context.connection_record and not oob_record:
            raise HandlerException(
                "No connection or associated connectionless exchange found for credential"
                " offer"
            )

        connection_id = (
            context.connection_record.connection_id if context.connection_record else None
        )

        credential_manager = CredentialManager(profile)
        cred_ex_record = await credential_manager.receive_offer(
            context.message, connection_id
        )  # mgr only finds, saves record: on exception, saving state null is hopeless

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="CredentialOfferHandler.handle.END",
            perf_counter=r_time,
        )

        if context.connection_record:
            holder_did = context.connection_record.my_did
        else:
            # Transform recipient key into did
            holder_did = default_did_from_verkey(oob_record.our_recipient_key)

        # If auto respond is turned on, automatically reply with credential request
        if cred_ex_record and context.settings.get("debug.auto_respond_credential_offer"):
            credential_request_message = None
            try:
                (
                    _,
                    credential_request_message,
                ) = await credential_manager.create_request(
                    cred_ex_record=cred_ex_record,
                    holder_did=holder_did,
                )
                await responder.send_reply(credential_request_message)
            except (
                BaseModelError,
                CredentialManagerError,
                IndyHolderError,
                LedgerError,
                StorageError,
            ) as err:
                self._logger.exception("Error responding to credential offer")
                if cred_ex_record:
                    async with profile.session() as session:
                        await cred_ex_record.save_error_state(
                            session,
                            reason=err.roll_up,  # us: be specific
                        )
                    await responder.send_reply(
                        problem_report_for_record(
                            cred_ex_record,
                            ProblemReportReason.ISSUANCE_ABANDONED.value,  # them: vague
                        )
                    )

            trace_event(
                context.settings,
                credential_request_message,
                outcome="CredentialOfferHandler.handle.REQUEST",
                perf_counter=r_time,
            )



from unittest import IsolatedAsyncioTestCase

from ..domain_txn_handler import get_proof_nodes, prepare_for_state_read
from ..hasher import HexTreeHasher
from ..merkel_verifier import MerkleVerifier
from ..trie import SubTrie
from .test_data import (
    GET_ATTRIB_REPLY,
    GET_CLAIM_DEF_REPLY_A,
    GET_CLAIM_DEF_REPLY_B,
    GET_CLAIM_DEF_REPLY_INVALID,
    GET_NYM_REPLY,
    GET_REVOC_REG_DEF_REPLY_A,
    GET_REVOC_REG_DEF_REPLY_B,
    GET_REVOC_REG_DELTA_REPLY_A,
    GET_REVOC_REG_DELTA_REPLY_B,
    GET_REVOC_REG_DELTA_REPLY_C,
    GET_REVOC_REG_REPLY_A,
    GET_REVOC_REG_REPLY_B,
    GET_SCHEMA_REPLY_A,
    GET_SCHEMA_REPLY_B,
    RAW_HEX_LEAF,
    SHA256_AUDIT_PATH,
)


class TestSubTrie(IsolatedAsyncioTestCase):
    def test_get_setter_root_hash(self):
        test_trie = SubTrie()
        test_trie.root_hash = 530343892119126197
        assert test_trie.root_hash == 530343892119126197

    def test_get_blank_node(self):
        assert SubTrie._get_node_type(b"") == 0

    async def test_verify_spv_proof_catch_exception(self):
        assert not await SubTrie.verify_spv_proof(
            expected_value="test", proof_nodes="test"
        )


class TestMPTStateProofValidation(IsolatedAsyncioTestCase):
    async def test_validate_get_nym(self):
        reply = GET_NYM_REPLY
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply),
            expected_value=prepare_for_state_read(reply),
        )

    async def test_validate_get_attrib(self):
        reply = GET_ATTRIB_REPLY
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply),
            expected_value=prepare_for_state_read(reply),
        )

    async def test_validate_get_schema(self):
        reply_a = GET_SCHEMA_REPLY_A
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_a),
            expected_value=prepare_for_state_read(reply_a),
        )
        reply_b = GET_SCHEMA_REPLY_B
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_b),
            expected_value=prepare_for_state_read(reply_b),
        )

    async def test_validate_get_claim_def(self):
        reply_a = GET_CLAIM_DEF_REPLY_A
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_a),
            expected_value=prepare_for_state_read(reply_a),
        )
        reply_b = GET_CLAIM_DEF_REPLY_B
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_b),
            expected_value=prepare_for_state_read(reply_b),
        )
        reply_c = GET_CLAIM_DEF_REPLY_INVALID
        assert not await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_c),
            expected_value=prepare_for_state_read(reply_c),
        )

    async def test_validate_get_revoc_reg(self):
        reply_a = GET_REVOC_REG_REPLY_A
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_a),
            expected_value=prepare_for_state_read(reply_a),
        )
        reply_b = GET_REVOC_REG_REPLY_B
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_b),
            expected_value=prepare_for_state_read(reply_b),
        )

    async def test_validate_get_revoc_reg_def(self):
        reply_a = GET_REVOC_REG_DEF_REPLY_A
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_a),
            expected_value=prepare_for_state_read(reply_a),
        )
        reply_b = GET_REVOC_REG_DEF_REPLY_B
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_b),
            expected_value=prepare_for_state_read(reply_b),
        )

    async def test_validate_get_revoc_reg_delta(self):
        reply_a = GET_REVOC_REG_DELTA_REPLY_A
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_a),
            expected_value=prepare_for_state_read(reply_a),
        )
        reply_b = GET_REVOC_REG_DELTA_REPLY_C
        assert await SubTrie.verify_spv_proof(
            proof_nodes=get_proof_nodes(reply_b),
            expected_value=prepare_for_state_read(reply_b),
        )
        reply_c = GET_REVOC_REG_DELTA_REPLY_B
        proof_nodes = get_proof_nodes(reply_c)
        expected_values = prepare_for_state_read(reply_c)
        assert await SubTrie.verify_spv_proof(
            proof_nodes=proof_nodes[0],
            expected_value=expected_values[0],
        )
        assert await SubTrie.verify_spv_proof(
            proof_nodes=proof_nodes[1],
            expected_value=expected_values[1],
        )


class TestMerkleRootHashValidation(IsolatedAsyncioTestCase):
    async def test_verify_leaf_inclusion_x(self):
        merkle_verifier = MerkleVerifier(HexTreeHasher())
        leaf_index = 848049
        tree_size = 3630887
        expected_root_hash = (
            b"78316a05c9bcf14a3a4548f5b854a9adfcd46a4c034401b3ce7eb7ac2f1d0ecb"
        )
        assert (
            await merkle_verifier.calculate_root_hash(
                RAW_HEX_LEAF,
                leaf_index,
                SHA256_AUDIT_PATH[:],
                tree_size,
            )
            == expected_root_hash
        )



import pytest
import pytest_asyncio

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......utils.testing import create_test_profile
from ...messages.problem_report import RotateProblemReport
from .. import problem_report_handler as test_module

test_valid_rotate_request = {
    "to_did": "did:example:newdid",
}


@pytest_asyncio.fixture
async def request_context():
    yield RequestContext.test_context(await create_test_profile())


class TestProblemReportHandler:
    """Unit tests for ProblemReportHandler."""

    @pytest.mark.asyncio
    @mock.patch.object(test_module, "DIDRotateManager")
    async def test_handle(self, MockDIDRotateManager, request_context):
        MockDIDRotateManager.return_value.receive_problem_report = mock.CoroutineMock()

        request_context.message = RotateProblemReport()
        request_context.connection_record = mock.MagicMock()
        request_context.connection_ready = True

        handler = test_module.ProblemReportHandler()
        responder = MockResponder()
        await handler.handle(request_context, responder)

        MockDIDRotateManager.return_value.receive_problem_report.assert_called_once_with(
            request_context.message
        )
